//------------------------------------------------------------------------------
//~Author: Fabio Arciniegas A.
//~Name: IndentHandler
//~Description: Implement all the methods of a SAX 2.0 ContentHandler, 
// outputting the soruce document with improved formatting according with
//  provided indentation, filling, and attribute-handling parameters
//
//Version $Id: IndentHandler.java,v 1.2 2001/01/28 15:48:13 Default Exp $
//~Important: See distribution note at the end of this file.

//------------------------------------------------------------------------------
//             This code has been formated to fit in 80 columns
//------------------------------------------------------------------------------
package com.faa.sax.util;

  /** 
      Implement all the methods of a SAX 2.0 ContentHandler, outputting the 
      soruce document with improved formatting according with provided 
      indentation, filling, and attribute-handling parameters
      @see Indent
   */

// Classes
import java.io.             OutputStreamWriter;
import java.io.             PrintWriter;
import java.io.             PrintStream;
import org.xml.sax.helpers. DefaultHandler;
import org.xml.sax.helpers. AttributesImpl;

// Interfaces
import org.xml.sax. Attributes;
import org.xml.sax. SAXException;
import org.xml.sax. XMLReader;

// Exceptions
import java.io.     UnsupportedEncodingException;
import org.xml.sax. SAXParseException;

public class IndentHandler extends DefaultHandler
{

// ******************************************         Constructors / Destructors
  /** 
      Create a new Indentation Handler that will output the document according 
      with the specified format.
      @param newTabs whether to use tabs for the indentation (or spaces)
      @param newIndent the amount of indentation characters per level
      @param newFill auto-fill text at given column. No autofill if 0
      @param newLineOffset if using autofill, use this as left offset 
      @param newAlignWithStart if auto-filling, align running text with start
                               or end of the tag
      @param newAlignFirstLine if auto-filling, start of text in the following 
                               line
      @param newOut output stream
      @see Indent
   */
     public IndentHandler(boolean newTabs,int newIndent,int newFill,
			 int newAttributes, int newLineOffset, 
			 boolean newAlignFirstLine, PrintStream newOut)
    {
	if(newTabs)
	    indentChar = '\t';
	indent = newIndent;
	fill = newFill;
	attributes = newAttributes;
	lineOffset = newLineOffset;
	alignFirstLine = newAlignFirstLine; 
	alignText = newFill != 0;
	out = newOut;
    }

// ******************************************                   Document Handler

   /** 
      Start document. 
    */
    public void startDocument() 
    {
            out.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
            out.flush();
    }

    /** 
      Print indented start tag and organize attributes accordingly
      @param uri the ns part of the qname
      @param local the local part of the qname
      @param raw the complete, raw name
      @param atts the attributes for the element
   */
    public void startElement(String uri, String local, 
                             String raw, Attributes attrs) 
    {
	boolean startMixed = chars.length() != 0 && containsFirstLine; 
	      // used to detect whether we ar on a mixed element
	if(chars.length() != 0 && (!containsFirstLine || !alignFirstLine))
	    printCharacters();

	else if(chars.length() == 0) // not within a mixed element
	{
    	    out.print("\n");
	    indent(true);
	}

	else if(chars.length() != 0 && containsFirstLine && alignFirstLine)
	{
	    out.print("\n");
	    indent();
	    for(int j=0;j <lineOffset;j++)
		out.print(' ');
	    printCharacters();
	}


	
        out.print('<');
        out.print(raw);
        if (attrs != null) 
	{
            int len = attrs.getLength();
            for (int i = 0; i < len; i++) 
	    {
                out.print(' ');
                out.print(attrs.getQName(i));
                out.print("=\"");
                out.print(normalize(attrs.getValue(i)));
                out.print('"');
		if((i+1)%attributes == 0 && i+1 < len) // time to do a new line
		{
                    out.print('\n');
		    for(int j=0;j < raw.length() + level*indent + 1;j++)
			out.print(' '); //indent the attribute under the first
		}
            }
        }
        out.print('>');
        out.flush();
	containsFirstLine = true;
	if(mixed > 0 || startMixed) // is is mixed content
	{
    	    mixed++;
	    curChar += raw.length() + 2;
	}

    } 

   /** 
      Output the end element tag.
      @param uri of the qname
      @return local part of the qname
   */
    public void endElement(String uri, String local, String raw) 
    {
	if(mixed == 0)
	    chars = chars.trim();

 	if(chars.length() != 0 && (!containsFirstLine || !alignFirstLine))
	{
	    printCharacters();
	    if(alignText && mixed == 0)
	    {
		out.print('\n');
		indent();
	    }
	}

	else if(chars.length() == 0)
	{
    	    out.print("\n");
	    indent();
	}

	else if(chars.length() != 0 && containsFirstLine && alignFirstLine)
	{
	    if(mixed == 0)
	    {
		out.print("\n");
		indent();
		for(int j=0;j <lineOffset;j++)
		    out.print(' ');
	    }
	    printCharacters();
	    if(mixed == 0)
	    {
		out.print('\n');
		indent();
	    }
	}
	

        out.print("</");
        out.print(raw);
        out.print('>');
        out.flush();
	if(mixed == 0) 
	{
	    curChar = 0;
	    deIndent();	    
	}// no indentation in mixed content

	if(mixed > 0) // in mixed content
	{
	    mixed--;
    	    curChar += raw.length() + 3;
	}

	containsFirstLine = false;
    } 

  /** 
      print the characters on buffer, taking care of filling and indenting issues
   */
 
    public void printCharacters()
    {
	if(mixed == 0)
	    // trim the starting whitespace
	    chars = chars.trim();
	boolean force;
	for(int i=0;i<chars.length();i++)
	{
	    int phy = curChar + level*indent + lineOffset; //physical position
	    if(force = (fill > 0 && 
		       (chars.charAt(i) == ' ' || i == 0 )&&
                       (chars.indexOf(' ',i+1) + phy - i) > fill))
	    {
		if(chars.charAt(i) == ' ')
		    chars = chars.substring(i+1);
		else
		    chars = chars.substring(i);
		i = 0;
	    }

	    if(fill != 0 && (phy >= fill  || force))
	    {
		out.print('\n');
		if(alignText)
		{
		    indent(); 
		    for(int j=0;j <lineOffset;j++)
			out.print(' ');
		}
		curChar = 0;
	    }
	    out.print(chars.charAt(i));
	    curChar++;
	}
	chars = "";
    }
  /** 
      output the characters, taking care of auto-filling and indenting.
      @param ch[] the whole stream of characters
      @return start start of the valid stream
      @exception length number of valid characters in the string
   */
    public void characters(char ch[], int start, int length) 
    {
	// just accumulate it. the output will be performed on startElement
        // or endElement
        chars = chars + normalize(new String(ch, start, length));
    } 

  /** 
      Insert the processing instruction, indented.
      @param target PI Target
      @return data PI Data
   */
    public void processingInstruction(String target, String data) 
    {
	indent();
        out.print("<?");
        out.print(target);
        if (data != null && data.length() > 0) 
	{
            out.print(' ');
            out.print(data);
        }
        out.print("?>");
        out.flush();
    } 

    
   /** 
       Normalize the given string
      @param s The string to normalize
      @return A string with special characters escaped (e.g &lt; is &amp;lt;)
   */
    protected String normalize(String s) 
    {
        StringBuffer str = new StringBuffer();
        int len = (s != null) ? s.length() : 0;
        for (int i = 0; i < len; i++) 
	{
            char ch = s.charAt(i);
            switch (ch) {
                case '<': {
                    str.append("&lt;");
                    break;
                }
                case '>': {
                    str.append("&gt;");
                    break;
                }
                case '&': {
                    str.append("&amp;");
                    break;
                }
                case '"': {
                    str.append("&quot;");
                    break;
                }
                    // else, default append char
                default: {
                    str.append(ch);
                }
            }
        }

        return str.toString();

    } 

// ******************************************                         ErrorHandler
    // The following methods were taken from the SAX2Writer example of 
    // Xerces.

    /** 
	Warning. 
    */
    public void warning(SAXParseException ex) 
    {
        System.err.println("[Warning] "+
                           getLocationString(ex)+": "+
                           ex.getMessage());
    }

    /** 
	Error. 
    */
    public void error(SAXParseException ex) 
    {
        System.err.println("[Error] "+
                           getLocationString(ex)+": "+
                           ex.getMessage());
    }

    /** 
	Fatal error. 
    */
    public void fatalError(SAXParseException ex) throws SAXException 
    {
        System.err.println("[Fatal Error] "+
                           getLocationString(ex)+": "+
                           ex.getMessage());
        throw ex;
    }

    /** 
	Returns a string of the location. 
    */
    private String getLocationString(SAXParseException ex) 
    {
        StringBuffer str = new StringBuffer();

        String systemId = ex.getSystemId();
        if (systemId != null) {
            int index = systemId.lastIndexOf('/');
            if (index != -1) 
                systemId = systemId.substring(index + 1);
            str.append(systemId);
        }
        str.append(':');
        str.append(ex.getLineNumber());
        str.append(':');
        str.append(ex.getColumnNumber());

        return str.toString();

    } 

// ******************************************                         Indentation
  /** 
      Simply put as many spaces or tabs is necessary for this level
      @param adding if this is true, increase the level 
   */
    private void indent(boolean adding)
    {
	if(adding)
	    level++;
	for(int i=0;i < level*indent;i++)
	    out.print(indentChar);
    }

  /** 
      Indent without adding
   */
    private void indent()
    {
	indent(false);
    }

  /** 
      Decrease the level of indentation
   */
    private void deIndent()
    {
	level--;
    }
     

// ******************************************                          Atributes

    char indentChar = ' ';
    int indent = 0;
    int fill = 0;
    int attributes = 1;
    int lineOffset = 1;
    boolean alignFirstLine = false; 
    boolean alignText = true;
    PrintStream out = System.out;

    int curChar = 0; // current character in the printed line
    int level = -1; // Indentation level
    String chars = "";  // The character data for this element (so far)
    boolean containsFirstLine; // Indicates whether chars contain the first line
    int mixed = 0; // Indicates whether we are inside mixed content
}












